/*
 * Copyright 2009 Red Hat, Inc.
 * Red Hat licenses this file to you under the Apache License, version
 * 2.0 (the "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
 * implied.  See the License for the specific language governing
 * permissions and limitations under the License.
 */
package org.hornetq.jms.example;

import org.hornetq.common.example.HornetQExample;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

public class ClusteredStandaloneWithNoConsumersExample extends HornetQExample {
    private Reconciler reconciler;

    public static void main(final String[] args) {
        new ClusteredStandaloneWithNoConsumersExample().run(args);
    }

    public boolean runExample() throws Exception {

        reconciler = new Reconciler();

        QueueSender sender = new QueueSender(reconciler);
        QueueConsumer consumer = new QueueConsumer(reconciler);
        consumer.start();
        sender.start();
        //Allow some time for some messages to be sent and consumed
        Thread.sleep(10000L);

        System.out.println("Now going to kill server 1");
        killServer(1);

        //Alow some time for any connections to restablish
        //and allow more messages to be send and consumed.
        Thread.sleep(10000L);

        //Stop the sender and allow the consumer
        //to consumer any messages left on the queue
        sender.stop();
        Thread.sleep(5000L);

        //Make sure there are no duplicates and messages which have not been received by consumer
        if (!checkCounts()) {
            return false;
        }
        //Start the sender again
        System.out.println("Restarting sender");
        sender.start();

        //Killing the backup server
        System.out.println("Now going to kill server 0");
        killServer(0);
        Thread.sleep(10000L);

        sender.stop();
        Thread.sleep(5000L);

        //It is now possible that there are unreconciled messages
        //so lets take a count as it shouldn't get any worse at this stage
        int unreconciled = reconciler.getUnReconciledCount();
        //There should however be no duplicates
        if (0 != reconciler.getLikelyDuplicateCount()) {
            System.out.println("**************\n\n\n\n\nThere are likely duplicates\n\n\n\n\n******************");
            return false;
        }

        //Restart the sender
        System.out.println("Restarting sender and allowing messages to prcoess");
        sender.start();
        //Alow some more messages to be sent from producer to consumer
        Thread.sleep(10000L);

        System.out.println("Stopping sender");
        sender.stop();
        //Allow time for any messages on the queue to be consumed
        Thread.sleep(5000L);

        System.out.println("Checking for further unreconciled messages");
        //There should be no more unreconciled messages
        if (unreconciled != reconciler.getUnReconciledCount()) {
            System.out.println("**************\n\n\n\n\nThere are additional unreconciled messages\n\n\n\n\n******************");
            return false;
        }

        consumer.stop();
        Thread.sleep(2000L);

        //We will now attempt to start live/backup pair server1/server0 respectively
        //and allow them to join the cluster

        //Firstly sync the data directory to ensure no data is lost
        syncData(0, 1);
        //Delete the Kill file
        deleteKill(0);
        System.out.println("Restarting server 0");
        startServer(0);
        //Allow time for restart
        Thread.sleep(5000L);
        //Delete the Kill file
        deleteKill(1);
        System.out.println("Restarting server 1");
        startServer(1);

        //CURRENTLY ALWAYS FAILING AFTER ATTEMPTING TO RESTART LIVE SERVER

        //Alow time for server startup
        Thread.sleep(5000L);
        System.out.println("Restarting sending");

        //Allow time for some messages to be sent and consumed
        Thread.sleep(10000L);

        sender.stop();
        //Alow time for any remaining messages to be consumed without any being sent
        Thread.sleep(2000L);


        //Now as the server0/server1 pair has rejoined the cluster all messages should now have been sent and consumed
        //Therefore recheck that there are no duplicates or missing messages.
        if (!checkCounts()) {
            return false;
        }

        //Stop the consumer and end
        consumer.stop();
        Thread.sleep(2000L);


        return true;
    }

    private boolean checkCounts() {
        boolean returnValue = true;
        if (reconciler.getUnReconciledCount() != 0) {
            System.out.println("***************\n\n\n\n\n\nThere are " + reconciler.getUnReconciledCount() + " unprocessed messages\n\n\n\n\n\n***************");
            returnValue = false;
        }
        if (reconciler.getLikelyDuplicateCount() != 0) {
            System.out.println("***************\n\n\n\n\n\nThere are " + reconciler.getLikelyDuplicateCount() + " processed messages which are likely to be duplicates\n\n\n\n\n\n***************");
            returnValue = false;
        }
        return returnValue;
    }

    private void syncData(int fromServer, int toServer) {
        System.out.println("Syncing data from server " + fromServer + " to server " + toServer);
        File fromDataDir = new File("server" + fromServer + "/data");
        File toDataDir = new File("server" + toServer + "/data");

        try {
            copyFile(fromDataDir, toDataDir);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void copyFile(File from, File to) throws Exception {
        if (from.isDirectory()) {
            for (File file : from.listFiles()) {
                try {
                    File newFile = new File(to, file.getName());
                    if (!newFile.exists()) {
                        newFile.createNewFile();
                    }
                    copyFile(file, newFile);
                } catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        } else {
            FileInputStream instream = new FileInputStream(from);
            FileOutputStream outstream = new FileOutputStream(to);

            byte[] buffer = new byte[1024];
            int len;

            while ((len = instream.read(buffer)) > 0) {
                outstream.write(buffer, 0, len);
            }
            instream.close();
            outstream.close();
        }
    }

    private void deleteKill(int id) {
        File file = new File("server" + id + "/KILL_ME");
        file.delete();
    }


}

