package org.hornetq.jms.example;

import java.util.concurrent.*;
import javax.jms.*;
import org.hornetq.jms.client.*;

/**
 * @author Grant Little <grant.little@coreservices.com.au>
 */
public class QueueConsumer implements ExceptionListener, MessageListener {

    private HornetQConnectionFactory connectionFactory;
    private Destination destination;
    private Connection connection;
    private Session session;
    private MessageConsumer consumer;
    private ScheduledExecutorService service;
    private Future thread;
    private boolean enabled = true;
    private Reconciler reconciler;

    public QueueConsumer(Reconciler reconciler) {
        this.reconciler = reconciler;
    }

    public void start() {
        service = Executors.newScheduledThreadPool(1);
        thread =  service.schedule(new ConsumerThreadTask(), 0L, TimeUnit.SECONDS);
        enabled = true;
    }

    public void stop() {
        enabled = false;
        closeResources();
        thread.cancel(false);
        service.shutdownNow();
    }

    private class ConsumerThreadTask implements Runnable {
        public void run() {
            boolean startNewThread = false;
            createResources();
            try {
                while (enabled) {
                    Message message = consumer.receive(5000L);
                    if (null != message) {
                        QueueConsumer.this.onMessage(message);
                    }
                }
            } catch (Exception e) {
                System.out.println("Exception thrown, closing resources and scheduling new consumer");
                e.printStackTrace();
                startNewThread = true;
            }
            finally {
                closeResources();
            }
            if (startNewThread) {
                thread =  service.submit(new ConsumerThreadTask());
            }

        }
    }

    public Connection createConnection() {
        if (connection == null) {
            try {
                connection = connectionFactory.createConnection();
                connection.setExceptionListener(this);
                connection.start();
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return connection;
    }

    public Session createSession() {
        if (session == null) {
            try {
                session = connection.createSession(true, Session.AUTO_ACKNOWLEDGE);
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return session;
    }

    public MessageConsumer createConsumer() {
        if (consumer == null) {

            try {
                consumer = session.createConsumer(getDestination());
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return consumer;
    }

    public Destination getDestination() {
        if (null == destination) {
            destination = new HornetQQueue("ExampleQueue");
        }
        return destination;
    }

    public void onException(JMSException e) {
        //Test the connection to see if failover has occurred
        //If not close and re-establish
        try {
            System.out.println("*********************************\nQueueConsumer.onException");
            //We will try to create a queue browser on the current session to make sure it is still valid
            QueueBrowser browser = session.createBrowser((Queue)getDestination());
            Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            session.close();
            browser.close();
            System.out.println("QueueConsumer: JMSException thrown, but it appears the connection has been re-established or has failed over");
        } catch (JMSException e1) {
            System.out.println("Current session is invalid, will not close resources and attempt to reconnect on next send");
            thread.cancel(true);
            closeResources();
            thread =  service.submit(new ConsumerThreadTask());
        }
        System.out.println("*********************************");
    }

    private void closeResources() {
        Utils.closeConsumer(consumer);
        consumer = null;
        Utils.closeSession(session);
        session = null;
        Utils.closeConnection(connection);
        connection = null;
        connectionFactory = null;
    }

    private void createResources() {
        createConnectionFactory();
        createConnection();
        createSession();
        createConsumer();

    }

    public void onMessage(Message message) {
        TextMessage txtMessage = (TextMessage)message;
        try {
            String id = txtMessage.getStringProperty("_HQ_DUPL_ID");
            reconciler.removeMessage(id);
            System.out.println("Received message " + id);
        } catch (JMSException e) {
            e.printStackTrace();
        }
    }

    public void createConnectionFactory() {
        if (null == connectionFactory) {
            connectionFactory = new HornetQConnectionFactory("231.7.7.7", 9876);
            connectionFactory.setFailoverOnInitialConnection(true);
            connectionFactory.setFailoverOnServerShutdown(true);
            connectionFactory.setReconnectAttempts(3);
        }

    }

}
