package org.hornetq.jms.example;

import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.Future;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import javax.jms.*;
import org.hornetq.jms.client.*;

/**
 * @author Grant Little <grant.little@coreservices.com.au>
 */
public class QueueSender implements ExceptionListener {

    private HornetQConnectionFactory connectionFactory;
    private Connection connection;
    private Session session;
    private MessageProducer producer;
    private AtomicInteger integer = new AtomicInteger(0);
    private boolean enabled=false;
    private ScheduledExecutorService executionService;
    private Future future;
    private Reconciler reconciler;
    private Destination destination;

    public QueueSender(Reconciler reconciler) {
        this.reconciler = reconciler;
    }

    public void start() {
        executionService = Executors.newScheduledThreadPool(1);
        future = executionService.scheduleAtFixedRate(new Runnable() {
            public void run() {
                try {
                    sendMessage();
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }, 100, 1000, TimeUnit.MILLISECONDS);
    }

    public void stop() {
        future.cancel(true);
        executionService.shutdownNow();
    }

    public void sendMessage() throws Exception {
        if (null == connectionFactory) {
            createResources();
        }
        int id = integer.getAndIncrement();
        String msgId = String.valueOf(id);
        try {
            Message message = session.createTextMessage("another message");
            message.setStringProperty("_HQ_DUPL_ID", msgId);
            producer.send(message);
            //Must do this before the commit otherwise the
            //Consumer may actually attempt to retrieve it before we can add it
            //to the reconciler
            reconciler.addMessage(msgId);
            session.commit();

        } catch (Exception e) {
            reconciler.removeMessage(msgId);
            System.out.println("JMSException thrown, closing resources");
            closeResources();
        }
    }

    public void createConnectionFactory() {
        if (null == connectionFactory) {
            connectionFactory = new HornetQConnectionFactory("231.7.7.7", 9876);
            connectionFactory.setFailoverOnInitialConnection(true);
            connectionFactory.setFailoverOnServerShutdown(true);
            connectionFactory.setReconnectAttempts(3);
        }

    }

    public Connection createConnection() {
        if (connection == null) {
            try {
                connection = connectionFactory.createConnection();
                connection.setExceptionListener(this);
                connection.start();
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return connection;
    }

    public Session createSession() {
        if (session == null) {
            try {
                session = connection.createSession(true, Session.AUTO_ACKNOWLEDGE);
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return session;
    }

    public MessageProducer createProducer() {
        if (producer == null) {

            try {

                producer = session.createProducer(getDestination());
            } catch (JMSException e) {
                throw new RuntimeException(e);
            }
        }
        return producer;
    }

    public Destination getDestination() {
        if (null == destination) {
            destination = new HornetQQueue("ExampleQueue");
        }
        return destination;
    }

    public void onException(JMSException e) {
        System.out.println("*********************************\nQueueSender.onException");
        //Test the connection to see if failover has occurred
        //If not close and re-establish
        try {
            //We will try to create a queue browser on the current session to make sure it is still valid
            session.createBrowser((Queue)getDestination());
            Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
            System.out.println("QueueSender: JMSException thrown, but it appears the connection has been re-established or has failed over");
            session.close();
        } catch (JMSException e1) {
            System.out.println("Current session is invalid, will not close resources and attempt to reconnect on next send");
            closeResources();
        }
        System.out.println("*********************************");
    }

    private void closeResources() {
        Utils.closeProducer(producer);
        producer = null;
        Utils.closeSession(session);
        session = null;
        Utils.closeConnection(connection);
        connection = null;
        connectionFactory = null;
    }

    private void createResources() {
        createConnectionFactory();
        createConnection();
        createSession();
        createProducer();
    }

}
